/************************************************************************ 
 *   echo - echo argum,ents                                             *
 *                                                                      *
 *   Syntax:     ex touch;'[-c] [-a] [-m] [-v] fileist...'              *
 *                                                                      *
 *      -n      do not add new line                                     *
 *                                                                      *
 *  (c) copyright 1991 by David J. Walker                               *
 *                                                                      *
 *  This program and its source code may be freely used and distributed *
 *  as long as this copyright notice remains intace, and no commercial  *
 *  gain is made from the distribution.                                 *
 ************************************************************************/

/*----------------------------------------------------------------------------
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  15 Sep 90   v1.1    DJW   - Upgraded to support embedded escape sequence
 *                              Those recognised are (same as SysV):
 *                                   \b, \c \f, \n, \r, \t, \v, \0v
 *
 *  02 Apr 94   v1.2    DJW   - Minor changes to compile warning free when
 *                              using C68 release 4.13
 *
 *  05 Nov 94   v1.3    DJW   - Minor changes to compile warning free when
 *                              using C68 Release 4.14
 *---------------------------------------------------------------------------*/

char    _prog_name[] = "echo";
char    _copyright[] = " {c}1990, D.J.Walker";
char    _version[] = "v1.3 ";

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef QDOS
#include <qdos.h>
/*
 *  Options to tailor program to QDOS/C68 environment
 */
void    (* _consetup)() = consetup_title;
int     (* _cmdwildcard)() = cmdexpand;
#endif /* QDOS */

#define SIZE 1024
char buf[SIZE];
int count;
int nflag = 0;
int backflag = 0;
int value;

void collect (char *);

void main(argc, argv)
  int argc;
  char *argv[];
{
  register int i;

  if (argc > 1 && strcmp(argv[1],"-n") == 0) {
    nflag++;
    argc--;
    argv++;
  }
  for (i = 1; i < argc; i++) {
    collect(argv[i]);
    if (i < argc - 1) collect(" ");
  }

  if (nflag == 0) collect("\n");
  if (backflag > 1) buf[count++] = value;

  /* Print accumulated output. */
  if (count > 0) 
      (void)write(1, buf, count);
  exit(0);
}

void    collect(s)
char *s;
{

/* Collect characters.  For efficiency, write them in large chunks. */
  char c;

  while ((c = *s++) != 0) {
    switch (backflag) {
    case 0:
            if (c == '\\') {
                backflag++;
                continue;
            }
            break;
    case 1:
            switch (c) {
            case 'b':   c='\b';
                        break;
            case 'c':   nflag=1;
                        backflag=0;
                        continue;
            case 'n':   c='\n';
                        break;
            case 'r':   c='\r';
                        break;
            case 'f':   c='\f';
                        break;
            case 't':   c='\t';
                        break;
            case 'v':   c='\v';
                        break;
            case '0':   value=0;
                        backflag++;
                        continue;
            default:    break;
            }
    case 2:
    case 3:
            if ((c >= '0') && (c <= '7')) {
                value = (value<<3) + (c - '0');
                backflag++;
                continue;
            }
            buf[count++] = value;
            break;
    case 4:
            if ((c >= '0') && (c <= '7'))
                c = (value<<3) + (c - '0');
            else
                buf[count++] = value;
            break;
    }
    if (count >= (SIZE-1)) {
        (void) write(1, buf, count);
        count = 0;
    }
    buf[count++] = c;
    backflag = 0;
  }
}

