/*

 qerror_c

 Print a QDOS error message if _ioerr set, else calls perror().

 Date		Author		Comments
 ----		------		--------
 25/08/1991	P.J. Taylor	Original.
 15/09/1991	P.J. Taylor	Amended.

 1.04
 ----
 07/03/1998	A.D.Ives	Source code reformatted.
				Additional spaces between message and
				error code

*/

#include <stdio.h>
#include <errno.h>
#include <qdos.h>

static char *qdos_errlist[]= {
  "Not Complete"	, /* -1 */
  "Invalid Job" 	, /* -2 */
  "Out of Memory"	, /* -3 */
  "Out of Range"	, /* -4 */
  "Buffer Full" 	, /* -5 */
  "Channel not Found"	, /* -6 */
  "Not Found"		, /* -7 */
  "Already Exists"	, /* -8 */
  "In Use"		, /* -9 */
  "End of File" 	, /* -10 */
  "Drive Full"		, /* -11 */
  "Bad Name"		, /* -12 */
  "Xmit Error"		, /* -13 */
  "Format Failed"	, /* -14 */
  "Bad Parameter"	, /* -15 */
  "Bad Medium"		, /* -16 */
  "Error in Expression" , /* -17 */
  "Overflow"		, /* -18 */
  "Not Implemented Yet" , /* -19 */
  "Read Only"		, /* -20 */
  "Bad Line"		, /* -21 */
} ;

int qdos_nerr= (int)(sizeof(qdos_errlist) / sizeof(qdos_errlist[0])) ;

void qerror(char *msg) {

  /* msg	Message prefix or NULL		*/

  long err ;
  if ( errno >=0 )
    perror(msg);
  else {
    err= (EOSERR == errno) ? -_oserr : -errno ;
    if ( msg != NULL )
      fprintf(stderr, "%s : ", msg) ;
    if ( err > qdos_nerr )
      fprintf(stderr, "QDOS error %d.\n", errno) ;
    else
      fprintf(stderr, "%s.\n", qdos_errlist[err-1] ) ;
  }
}
