/*

 sh_c

 A Command Shell for QDOS & SMS.

 Date		Author		Comments
 ----		------		--------
 17/08/1991	P.J. Taylor	Original.
 10/10/1991	P.J. Taylor	Amended.
 26/02/1998	A.D. Ives	Reformatted code (put ifs 'right' way round).
				ANSI function headers adopted.
 27/02/1998	A.D.Ives	QUIT allowed as a synonym for EXIT.
				VER command added.
 01/03/1998	A.D.Ives	Modified the behaviour of the echo and prompt
				switches to make execution of secondary
				shells tidier. As a result, added the new ECHO
				command.
				Added customisable prompt.
				Added parameter substitution via SubstParams().
				Up to 26 parameters may be referenced as $A
				to $Z, environment variables via $(NAME), and
				system variables with $(%name)
 02/03/1998	A.D.Ives	When there has been no match from the list of
				internal commands, now tries calling
				FileInfo II to execute a data file.
 04/03/1998	A.D.Ives	Added recognition of the shorthand CD commands
				to change up one level, and go to root
				(CD\ and CD..) - see change to shint_c to
				support these parameters.
 06/03/1998	A.D.Ives	Added MD command to make a directory.
				Format of error messages changed.

 1.04
 ----
 07/03/1998	A.D.Ives	New %retcode variable added - holds the value
				of the last Return Code from loading a Job.
				Internal command execution moved out of main()
				into separate Obey() function.
				Added storage of last command line, and the
				ability to re-execute by starting a line
				with an '=', or to interactively edit the
				last line by starting with '/'. Lines that
				begin with a '#' are now treated as zero
				length (i.e. comments).
				Added PAUSE command.
				Added very simple non-nestable IF handling.
 1.05
 ----
 09/03/1998	A.D.Ives	CD, CPD, CDD now all call the enhanced cd()
				routine, passing a flag to indicate which
				directory is to be changed. Also check for
				@ as start of dir string (to support setting
				the OS Dir as well).
				New function CanExec() to test whether the
				Shell knows how to execute a command.
				Command execution changed so that it does
				this test first of all (stops calls to
				FileInfo II for explicitly stated _bat files,
				for example!)
				Added support for LET command to set the
				value of a local environment variable.
				Simple PRINT command added.
				Added support for OPT command to set switches
				in the new SHELL_OPTS environment variable.

 1.06
 ----
 10/03/1998	A.D.Ives	Common routines moved out into libpjt.
				New facilities for directory navigation
				on filenames used as parameters for
				internal commands, and as external
				commands ("\" ".\" "..\" prefixes)
				Modified some tests to support "/" as a
				synonym for "\".
				New Shell options +u and +l to force case
				on directory names.
				New internal variable "%lastkey" holds the
				ASCII code of the key that was hit to exit a
				PAUSE.
				Last command line only stored if it originated
				from the CONsole.
				Fixed bug that caused the original prompt
				string to be overwritten by the translated
				version.
				Added new prompt characters %D and %T for
				Date and Time.
				Added new internal variables %date and
				%time.
 11/03/1998			Prompt(), SubstParams() moved to new
				module suppt_c.
				Shell Internals no longer maintained as
				globals, but stored as statics in the
				suppt module, accessed via a Get/Set call.
				Fixed missing format string in fprint() call
				when printing prompt string that mean % signs
				were interpreted as formatting symbols (just
				as well there was no way to get a % into the
				prompt string due to another bug!)
 12/03/1998			Added override for default CONsole setup
				so that I could move the Y origin just
				slightly to stop it blocking my MultiButton
				window.
				New command CON to set all window options in
				one hit. Dropped variables that were storing
				the border and origin details as calls to
				winpos(), winsize(), and border() all route
				to SetCon() now.

 1.07
 ----
 14/03/1998			Pulled out all of the code relating to
				getting and processing the command line
				into a separate Process() function - which
				can now be called with a single string
				argument to execute any command.
				FOR command added.
 15/03/1998			PROGRAM_VERSION constant moved into sh_h
 21/03/1998			Tests for the presence of an '@' sign at the
				start of the command line and, if present,
				executes the command line immediately and
				exits.

 1.08
 ----
 04/04/1998			DO command added.
				New environment variable SHELL_PROMPT can
				be used to specify the prompt on entry to
				the Shell.

*/

char _PROG_NAME[]= "Sh" ;

#include <stdio.h>
#include <stdlib.h>
#include <qdos.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <debug.h>
#include <unistd.h>
#include <time.h>
#include <qptr.h>
#include "sh.h"
#include "shexec.h"
#include "shint.h"
#include "suppt.h"
#include "libpjt.h"

/* Return value from getcmds() ...  */

#define CMD_OK	0		/* No problems			*/
#define CMD_SYN 1		/* User syntax error		*/
#define CMD_EOF 2		/* End of file reached		*/

static
char PromptString[64] , 	/* Current prompt string	*/
     LastLine[256] ;		/* Last command line		*/

char **argStrings ;		/* Program arguments		*/

static
int  IsPrompt , 		/* Display a prompt		*/
     IsRunning ,		/* Keep running 		*/
     IsEcho,			/* Echo to stdout required	*/
     InIsCon ,			/* Input is from console	*/
     OpIsCon ,			/* Output is to console 	*/
     InIF ,			/* TRUE if inside an IF 	*/
     ExitCode ; 		/* Shell Exit Code		*/

int  argCount ; 		/* Program argument count	*/

static
CMDTAB ct ;			/* Table of commands		*/

void ConSetup(chanid_t chanid, struct WINDOWDEF *condetails) {

  /* This routine overrides the C68 default CONsole setup routine */

  SetCon(chanid, NULL) ;

}

void (* _consetup) () = ConSetup ;

char *getline(char *lbuff) {

  /*

  Get a command line. Returns the address of the first non-blank
  character in lbuff, or NULL.

  lbuff 	Input Line buffer

  */

  int len, retcode ;
  short elen ;
  char *res, ExitChar, pr[80] ;

  if ( IsPrompt ) {
    Prompt(PromptString, pr) ;
    fprintf(stdout, "%s", pr) ;
    fflush(stdout) ;
  }
  if ( (res= fgets(lbuff, 256, stdin)) == NULL )
    return res ;
  if ( (len= strlen(lbuff)) > 0 )
    lbuff[len - 1]= '\0' ;
  if ( IsEcho )
    printf("%s\n", lbuff) ;
  while ( *res == ' ' || *res == '\t' )
    res++ ;

  if ( InIsCon ) {
    if ( strcmp(res, "=") == 0 )  /* Use a copy of the last command */
      res= strcpy(lbuff, LastLine) ;
    else
      if ( strcmp(res, "/") == 0 ) {
	strcpy(lbuff, LastLine) ;
	len= strlen(lbuff) ;
	elen= (short)(len) ;
	res= lbuff + len ; /* Horrible isn't it? But that's what io_edlin wants! */
	retcode= io_edlin(fgetchid(stdin), -1, &res, 256, 0, &elen) ;
	if ( retcode == 0 )
	  ExitChar= *(res - 1) ;
	if ( retcode == 0 && ExitChar != 27 ) {     /* Returned OK, and not [ESC] */
	  *(res - 1)= '\0' ;
	  strcpy(LastLine, lbuff) ;
	}
	else
	  lbuff[0]= '\0' ;
	res= lbuff ;
      }
      else
	strcpy(LastLine, res) ;
  }

  if ( res[0] == '#' )
    res[0]= '\0' ;	/* Ignore a commented line completely */
  else
    if ( InIF ) {       /* Easiest to handle changed IF state right here */
      if ( (strnicmp(res, "else", 4) == 0) || (strnicmp(res, "endif", 5) == 0 ) )
	InIF= FALSE ;	/* ELSE or ENDIF clear the flag */
      res[0]= '\0' ;	/* Cancel the line to prevent double processing */
    }

  return res ;
}

char *nextcmd(char *cmd, int *xopt) {

  /*

  Returns the start of the next command and the execution option for
  the current command.

  cmd		Current command
  xopt		Execution option

  */

  int escd ;	   /* double '\' found */
  char *res ,
       *p ,	   /* current scan position less one */
       lq ;	   /* last quote char while in quotes else 0 */

  *xopt= EXE_WAIT ;
  res= cmd + strlen(cmd) ;
  for ( p= cmd, lq= 0, escd= FALSE ; p[1] ; p++) {
    if ( *p == '\\' && !escd )
      ;
    else
    if ( lq && (lq) == p[1]  )
      lq= 0 ;
    else
    if ( !lq && ( p[1] == '\'' || p[1] == '"' ) )
      lq= p[1] ;
    else
    if ( lq )
      ;
    else
    if ( p[1] == ';' ) {
      p[1]= 0 ;
      res= p + 2 ;
      break ;
    }
    else
    if ( *p != '>' && p[1] == '&' ) {
      p[1]= 0 ;
      res= p + 2 ;
      *xopt= EXE_FORK ;
      break ;
    }
    else
    if ( p[1] == '|' ) {
      p[1]= 0 ;
      if ( p[2] == '&' ) {
	res= p + 3 ;
	*xopt= EXE_EPIPE ;
      }
      else {
	res= p + 2 ;
	*xopt= EXE_PIPE ;
      }
      break ;
    }
    escd= ( *p == '\\' && p[1] == '\\' ) ;
  }
  while ( *res == ' ' || *res == '\t' )
    res++ ;
  if ( !*res )
  { if ( lq ) {
      fprintf(stderr, "Syntax Error: \"%c\" expected.\n", lq) ;
      *xopt= EXE_ERR ;
    }
    else
    if ( *xopt == EXE_PIPE || *xopt == EXE_EPIPE ) {
      fprintf(stderr, "Syntax Error: \"|\" found at end of line.\n") ;
      *xopt= EXE_ERR ;
    }
  }
  return res ;
}

char *separg(char *cmd) {

  /*

  Return pointer to arguments, and terminate the command with a \0.

  cmd		Current command (\0 terminator added)

  */

  int offs ;
  char *res ;

  res= cmd ;
  while ( *res && *res != ' ' && *res != '\t' )
    res++ ;
  if ( *res ) {
    *res++ = 0 ;
    while ( *res == ' ' || *res == '\t' )
      res++ ;
  }
  return res ;
}

int getcmds(char *lbuff) {

  /*

    Get commands from a string into the Command Table.

    Pass a zero length string to get a line from the user to process first.

  */

  char *ncmd ,		/* Pointer to next command		*/
       temp[8] ;	/* Temporary 'lookahead' buffer 	*/

  ct.n= 0 ;
  if ( *lbuff )
    ncmd= lbuff ;
  else
    if ( (ncmd= getline(lbuff) ) == NULL )
      return CMD_EOF ;

  /* Peek ahead at the start of the command line */

  strncpy(temp, lbuff, 4) ;
  temp[4]= 0 ;
  strlwr(temp) ;

  if ( (strcmp(temp, "for ") != 0) && (strncmp(temp, "do ", 3) !=0) )

    /* Inhibits substitution for an internal FOR command */

    SubstParams(lbuff, argCount, argStrings) ;

  while ( ct.n < NCMDS && *ncmd ) {
    ct.cmd[ct.n].name= ncmd ;
    ncmd= nextcmd(ncmd, &(ct.cmd[ct.n].xopt)) ;
    if ( ct.cmd[ct.n].xopt == EXE_ERR )
      return CMD_SYN ;
    ct.cmd[ct.n].arg= separg( ct.cmd[ct.n].name ) ;
    ct.n++ ;
  }
  if ( *ncmd ) {
    fprintf(stderr, "Too many commands on the same line, maximum=%d\n", NCMDS) ;
    return CMD_SYN ;
  }
  else
    return CMD_OK ;
}

int Obey(char *name, char *arg) {

  /* Identify and obey an internal command */

  int ret= TRUE ;

  if ( stricmp(name, "if") == 0 )
    InIF= !IfTest(arg) ; /* Setting InIF TRUE causes lines to be skipped */
  else
  if ( stricmp(name, "else") == 0 )
    InIF= !InIF ;
  else
  if ( stricmp(name, "endif") == 0 )
    InIF= FALSE ;
  else
  if ( stricmp(name, "ver") == 0 && OpIsCon )
    printf("Shell v%s\n", PROG_VERSION) ;
  else
  if ( /* Check for "echo on" or "echo off" to avoid clash with external echo */
       (stricmp(name, "echo") == 0 && OpIsCon) &&
       ( (stricmp(arg, "on") == 0) ||
	 (stricmp(arg, "off") == 0)
	)
     )
    SwitchEcho(arg, &IsEcho) ;
  else
  if ( stricmp(name, "prompt") == 0 && OpIsCon )
    SetPrompt(arg, PromptString) ;
  else
  if ( stricmp(name, "let") == 0 )
    let(arg) ;
  else
  if ( stricmp(name, "opt") == 0 )
    opt(arg) ;
  else
  if ( stricmp(name, "print") == 0 )
    print(arg) ;
  else
  if ( stricmp(name, "for") == 0 )
    forcmd(arg) ;
  else
  if ( stricmp(name, "do") == 0 )
    docmd(arg) ;
  else
  if ( stricmp(name, "con") == 0 && OpIsCon )
    SetCon(fgetchid(stdout), arg) ;
  else
  if ( stricmp(name, "cls") == 0 && OpIsCon )
    cls() ;
  else
  if ( stricmp(name, "ink") == 0 && OpIsCon )
    ink(arg) ;
  else
  if ( stricmp(name, "paper") == 0 && OpIsCon )
    paper(arg) ;
  else
  if ( stricmp(name, "winpos") == 0 && OpIsCon )
    winpos(arg) ;
  else
  if ( stricmp(name, "winsize") == 0 && OpIsCon )
    winsize(arg) ;
  else
  if ( stricmp(name, "border") == 0 && OpIsCon )
    border(arg) ;
  else
  if ( stricmp(name, "wait") == 0 )
    waitabit(arg) ;
  else
  if ( stricmp(name, "kill") == 0 )
    kill(arg) ;
  else
  if ( stricmp(name, "cd") == 0 )
    cd(arg, D_DATAD) ;
  else
  if ( (strnicmp(name, "cd\\", 3) == 0) ||
       (strnicmp(name, "cd/", 3)  == 0) ||
       (strnicmp(name, "cd.", 3)  == 0) ||
       (strnicmp(name, "cd@", 3)  == 0)
     )
    cd(name + 2, D_DATAD) ;
  else
  if ( stricmp(name, "cpd") == 0 )
    cd(arg, D_PROGD) ;
  else
  if ( (strnicmp(name, "cpd\\", 4) == 0) ||
       (strnicmp(name, "cpd/", 4) == 0)  ||
       (strnicmp(name, "cpd.", 4) == 0)  ||
       (strnicmp(name, "cpd@", 4) == 0)
     )
    cd(name + 3, D_PROGD) ;
  else
  if ( stricmp(name, "cdd") == 0 )
    cd(arg, D_DESTD) ;
  else
  if ( (strnicmp(name, "cdd\\", 4) == 0) ||
       (strnicmp(name, "cdd/", 4) == 0)  ||
       (strnicmp(name, "cdd.", 4) == 0)  ||
       (strnicmp(name, "cdd@", 4) == 0)
     )
    cd(name + 3, D_DESTD) ;
  else
  if ( stricmp(name, "md") == 0 )
    md(arg) ;
  else
  if ( (strnicmp(name, "md\\", 3) == 0) ||
       (strnicmp(name, "md/", 3) == 0)	||
       (strnicmp(name, "md.", 3) == 0)
     )
    md(name + 2) ;
  else
  if ( stricmp(name, "pause") == 0 )
    WaitKey(arg) ;
  else
  if ( stricmp(name, "ftest") == 0 )
    FTest(arg) ;
  else
    ret= FALSE ;

  return ret ;

}

int CanExec(char *name) {

  /* Does the Shell know how to execute a file? */

  int ExtenPos, res ;
  char *ExtenPtr ;

  res= FALSE ;
  ExtenPos= strrpos(name, '_') ;
  if (ExtenPos != -1) {
    ExtenPtr= name + ExtenPos ;
    if ( strcmpi(ExtenPtr, "_bat") == 0 )
      res= TRUE ;
    else
    if ( stricmp(ExtenPtr, "_obj") == 0 )
      res= TRUE ;
    else
    if ( stricmp(ExtenPtr, "_exe") == 0 )
      res= TRUE ;
    else
    if ( stricmp(ExtenPtr, "_rex") == 0 )
      res= TRUE ;
  }
  return res ;
}

int Process(char *lbuff) {

  /*

    Pass a string to be processed as a command.

    Pass a zero length string to get a line from the user and
    then process it.

  */

  int res ;

  res= getcmds(lbuff) ;

  if ( res != CMD_EOF ) {
    if ( ct.n == 1 && ct.cmd[0].xopt == EXE_WAIT ) {
      if ( CanExec(ct.cmd[0].name) )
	execcmds(&ct, OpIsCon) ;
      else {
	if ( ( stricmp(ct.cmd[0].name, "exit") == 0 ) ||
	     ( stricmp(ct.cmd[0].name, "quit") == 0 ) ) {
	  if ( ct.cmd[0].arg != NULL )
	    ExitCode= atoi(ct.cmd[0].arg) ;
	  IsRunning= FALSE ;
	}
	else {
	  if ( !Obey(ct.cmd[0].name, ct.cmd[0].arg) )
	    if ( !fexec(ct.cmd[0].name) )
	      execcmds(&ct, OpIsCon) ;
	}
      }
    }
    else
      execcmds(&ct, OpIsCon) ;
  }

  return res ;
}

main(int argc, char **argv) {

  int res ;
  char lbuff[256] ,	/* Input line buffer			*/
       *mprompt ;

  /* Initialise Flags */

  ShellInit() ;
  InIsCon= isatty(fileno(stdin)) ;
  OpIsCon= isatty(fileno(stdout)) ;
  IsPrompt= (InIsCon && OpIsCon) ;
  IsEcho= FALSE ;
  IsRunning= TRUE ;
  InIF= FALSE ;
  argCount= argc ;
  argStrings= argv ;
  mprompt= getenv("SHELL_PROMPT") ;
  if ( mprompt == NULL )
    strcpy(PromptString, ">") ;
  else
    strcpy(PromptString, mprompt) ;
  LastLine[0]= '\0' ;
  _endmsg= NULL ;
  res= CMD_OK ;
  if ( argc > 1 && *argv[1] == '@' ) {

    /*

      If the first character on the command line is an '@' sign
      rebuild a line to be parsed from the parameters, and
      execute it straight away.

    */

    int i ;

    strcpy(lbuff, argv[1] + 1) ;
    for ( i= 2 ; i <= argc ; i++ ) {
      strcat(lbuff, " ") ;
      strcat(lbuff, argv[i]) ;
    }
    res= Process(lbuff) ;
  }
  else
    while ( IsRunning && res != CMD_EOF && !feof(stdin) ) {
      *lbuff= 0 ;
      res= Process(lbuff) ;
    }
  return ExitCode ;
}
